import os
from pyxnat import Interface
from requests.exceptions import ConnectionError
import os.path as op
from nose import SkipTest
from functools import wraps
from nose.plugins.attrib import attr


fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')

def docker_available(func=None):
    '''Skip test completely if no Docker-based XNAT instance available
    '''

    def check_and_raise():
        x = Interface(config='.xnat.cfg')
        try:
            x.head('')
            projects = list(x.select.projects())
            print('Docker instance found.')
        except (ConnectionError, KeyError):
            print('Skipping it.')
            raise SkipTest('Docker-based XNAT instance unavailable')

    if func:
        @wraps(func)
        @attr('network')
        @attr('docker_available')
        def newfunc(*args, **kwargs):
            check_and_raise()
            return func(*args, **kwargs)
        return newfunc
    else:
        check_and_raise()

@docker_available
def test_users():
    x = Interface(config='.xnat.cfg')
    assert isinstance(x.manage.users(), list)

@docker_available
def test_user_firstname():
    x = Interface(config='.xnat.cfg')
    print(x._get_json('%s/users' % x._entry))
    assert x.manage.users.firstname('admin') == 'Admin'

@docker_available
def test_user_lastname():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.lastname('admin') == 'Admin'

@docker_available
def test_user_email():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.email('admin') == \
        'fake@fake.fake'

@docker_available
def test_user_id():
    x = Interface(config='.xnat.cfg')
    assert x.manage.users.id('admin') == '1'

@docker_available
def test_add_remove_user():
    x = Interface(config='.xnat.cfg')
    x.select.project('nosetests3').remove_user('admin')
    x.select.project('nosetests3').add_user('admin', 'collaborator')
    assert 'admin' in x.select.project('nosetests3').collaborators()
    x.select.project('nosetests3').remove_user('admin')
    assert 'admin' not in x.select.project('nosetests3'
                                                    ).collaborators()
    x.select.project('nosetests3').add_user('admin', 'owner')

@docker_available
def test_project_accessibility():
    x = Interface(config='.xnat.cfg')
    print(x.select.project('nosetests3').accessibility())
    assert x.select.project('nosetests3').accessibility() in \
                                        [b'public', b'protected', b'private']
    x.select.project('nosetests3').set_accessibility('private')
    assert x.select.project('nosetests3').accessibility() == b'private'
    x.select.project('nosetests3').set_accessibility('protected')
    assert x.select.project('nosetests3').accessibility() == b'protected'

@docker_available
def test_create_xml():
    x = Interface(config='.xnat.cfg')
    _modulepath = op.dirname(op.abspath(__file__))
    p = x.select.project('nosetests')
    s = p.subject('10001')
    e = s.experiment('10001_MR')
    fp = op.join(_modulepath, 'sess.xml')
    assert(op.isfile(fp))
    e.create(xml=fp)
    xml = e.get().decode()
    assert('Alomar' in xml)

def test_project_users():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests3').users(), list)

def test_project_owners():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests3').owners(), list)

def test_project_members():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests3').members(), list)

def test_project_collaborators():
    x = Interface(config=fp)
    assert isinstance(x.select.project('nosetests3').collaborators(),
                      list)

def test_project_user_role():
    x = Interface(config=fp)
    assert x.select.project('nosetests3').user_role('nosetests') == 'owner'



from six import string_types
# REST collection resources tree
resources_tree = {
    'projects'        : ['subjects', 'resources'],
    'subjects'        : ['experiments', 'resources'],
    'experiments'     : ['assessors', 'reconstructions', 'scans', 'resources'],
    'assessors'       : ['resources', 'in_resources','out_resources'],
    'reconstructions' : ['in_resources','out_resources'],
    'scans'           : ['resources'],
    'resources'       : ['files'],
    'files'           : [],
    'in_resources'    : ['files'],
    'in_files'        : [],
    'out_resources'   : ['files'],
    'out_files'       : [],
    }

prearc_tree = {'projects'   : ['scans'],
               'scans'      : ['resources'],
               'resources'  : ['files']
               }
# REST resources that are not natively supported
extra_resources_tree = {'projects' : ['assessors', 'scans', 'reconstructions'],
                        'subjects' : ['assessors', 'scans', 'reconstructions'],
                        }

# REST <Python to URI> translation table
rest_translation = {'in_resources' : 'in/resources',
                    'in_files' : 'in/files',
                    'out_resources' : 'out/resources',
                    'out_files' : 'out/files',
                    'in_resource' : 'in/resource',
                    'in_file' : 'in/file',
                    'out_resource' : 'out/resource',
                    'out_file' : 'out/file',
                    }



# REST json format <id_header, label_header>
json = {'projects' : ['ID', 'ID'],
        'subjects' : ['ID', 'label'],
        'experiments' : ['ID', 'label'],
        'assessors' : ['ID', 'label'],
        'reconstructions' : ['ID', 'label'],
        'scans' : ['ID', 'ID'],
        'resources' : ['xnat_abstractresource_id', 'label'],
        'out_resources' : ['xnat_abstractresource_id', 'label'],
        'in_resources' : ['xnat_abstractresource_id', 'label'],
        'files' : ['path', 'path'],
        }

resources_singular = [key.rsplit('s', 1)[0] for key in resources_tree.keys()]
resources_plural   = resources_tree.keys()
resources_types    = resources_singular + list(resources_plural)

default_datatypes = {'projects' : 'xnat:projectData',
                     'subjects' : 'xnat:subjectData',
                     'experiments' : 'xnat:mrSessionData',
                     'assessors' : 'xnat:mrAssessorData',
                     'reconstructions' : 'xnat:reconstructedImageData',
                     'scans' : 'xnat:mrScanData',
                     'resources' : None,
                     'in_resources' : None,
                     'out_resources' : None,
                     }

def datatype_attributes(root, datatype):
    def _iterchildren(node, pathsofar):
        elements = []

        for child in node.iterchildren():
            if isinstance(child.tag, string_types) \
                and child.tag.split('}')[1] == 'element':
                    elements.append('%s/%s'%(pathsofar, child.get('name')))
                    elements.extend(_iterchildren(child, '%s/%s'%
                                        (pathsofar, child.get('name')))
                                    )

            elif isinstance(child.tag, string_types) \
                and child.tag.split('}')[1] == 'attribute':
                    elements.append('%s/%s'%(pathsofar, child.get('name')))

            elif isinstance(child.tag, string_types) \
                    and child.tag.split('}')[1] == 'extension':

                ct_xpath = "/xs:schema/xs:complexType[@name='%s']"% \
                    child.get('base').split(':')[1]

                for complex_type in node.getroottree(
                    ).xpath(ct_xpath, namespaces=child.nsmap):

                    same = False

                    for ancestor in child.iterancestors():
                        if ancestor.get('name') == \
                                child.get('base').split(':')[1]:
                            same = True
                            break

                    if not same:
                        elements.extend(_iterchildren(complex_type,
                                                      pathsofar)
                                        )

                    elements.extend(_iterchildren(child, pathsofar))

            else:
                elements.extend(_iterchildren(child, pathsofar))

        return elements

    ct_xpath = "/xs:schema/xs:complexType[@name='%s']" % \
        datatype.split(':')[1]

    attributes = []

    for complex_type in root.xpath(ct_xpath, namespaces=root.nsmap):
        for child in complex_type.iterchildren():
            attributes.extend(_iterchildren(child, datatype))

    return attributes

def datatypes(root):
    nsmap = get_nsmap(root)

    return [element.get('type')
            for element in root.xpath('/xs:schema/xs:element',
                                      namespaces=nsmap)
            ]

def get_nsmap(node):
    nsmap = node.nsmap
    none_ns = node.nsmap.get(None)

    if none_ns != None:
        nsmap[none_ns.rsplit('/', 1)[1]] = none_ns
        del nsmap[None]

    return nsmap

def class_name(self):
    """
    Return the name of this class without qualification.
    eg. If the class name is "x.y.class" return only "class"
    """
    return self.__class__.__name__.split('.')[-1]



import sys
import os.path as op
import pyxnat
from . import skip_if_no_network

_modulepath = op.dirname(op.abspath(pyxnat.__file__))

dd = op.join(op.split(_modulepath)[0], 'bin')
sys.path.append(dd)

@skip_if_no_network
def test_001_sessionmirror():
    from sessionmirror import create_parser, main
    parser = create_parser()
    cfg = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
    central = pyxnat.Interface(config=cfg)
    e = 'CENTRAL_E74609'
    args = ['--h1', cfg, '--h2', cfg, '-e', e, '-p', 'nosetests3']
    args = parser.parse_args(args)
    main(args)
    e1 = central.array.experiments(experiment_id=e,
                                  columns=['subject_label']).data[0]

@skip_if_no_network
def test_002_deletesubject():
    print('DELETING')
    cfg = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
    central = pyxnat.Interface(config=cfg)
    e = 'CENTRAL_E74609'
    e0 = central.array.experiments(experiment_id=e,
        columns=['subject_label', 'label']).data[0]
    subject_label = e0['subject_label']
    experiment_label = e0['label']
    e1 = central.array.experiments(project_id='nosetests3',
                                subject_label=subject_label,
                                experiment_label=experiment_label,
                                columns=['subject_id']).data[0]
    e2 = central.select.project('nosetests3').subject(e1['subject_ID']).experiment(e1['ID'])
    assert(e2.exists())
    e2.delete()
    assert(not e2.exists())



XNAT_RESOURCE_NAME = 'DICOM'

def scandate(self):
    import os.path as op
    import os
    import tempfile
    import pydicom
    from datetime import datetime

    f = list(self.files())[0]

    fp = op.join(tempfile.gettempdir(), 'test.dcm')
    _, fp = tempfile.mkstemp(suffix='.dcm')
    f.get(dest=fp)
    d = pydicom.read_file(fp)

    if hasattr(d, 'AcquisitionDate'):
        acquisition_date = d.AcquisitionDate
    else:
        acquisition_date = d.AcquisitionDateTime[:8]

    os.remove(fp)
    ad = datetime.strptime(acquisition_date, '%Y%m%d').strftime('%m-%d-%Y')
    return ad



import unittest
import os.path as op
from pyxnat import Interface
from . import skip_if_no_network
from nose import SkipTest


import logging as log
log.basicConfig(level=log.INFO)

class ArrayTests(unittest.TestCase):
    ''' Resource-related XNAT connectivity test cases '''

    def setUp(self):
        self._interface = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

    @skip_if_no_network
    def test_array_experiments(self):
        '''
        Get a list of experiments from a given subject which has multiple types
        of experiments (i.e. MRSessions and PETSessions) and assert it gathers
        them all.
        '''
        e = self._interface.array.experiments(subject_id='CENTRAL_S06242').data
        self.assertGreaterEqual(len(e), 3)

    @skip_if_no_network
    def test_array_mrsessions(self):
        '''
        From a given subject which has multiple types of experiments, get a list
        of MRI sessions and assert its length matches the list of experiments of
        type 'xnat:mrSessionData'
        '''
        mris = self._interface.array.mrsessions(subject_id='CENTRAL_S06242').data
        exps = self._interface.array.experiments(subject_id='CENTRAL_S06242',
                                                 experiment_type='xnat:mrSessionData'
                                                 ).data
        self.assertListEqual(mris, exps)

    @skip_if_no_network
    def test_array_scans(self):
        '''
        Get a list of scans from a given experiment which has multiple types
        of scans (i.e. PETScans and CTScans) and assert it gathers them all.
        '''
        s = self._interface.array.scans(experiment_id='CENTRAL_E72012').data
        self.assertEqual(len(s),16)

    @skip_if_no_network
    def test_array_mrscans(self):
        '''
        Get a list of MRI scans from a given experiment which has multiple scans
        mixed (i.e. MRScans and MRSpectroscopies, aka OtherDicomScans) and
        assert its length matches the list of scans filtered by type 'xnat:mrScanData'
        '''
        mris = self._interface.array.mrscans(experiment_id='CENTRAL_E72012').data
        exps = self._interface.array.scans(experiment_id='CENTRAL_E72012',
                                           scan_type='xnat:mrScanData'
                                           ).data
        self.assertListEqual([i['xnat:mrscandata/id'] for i in mris],
                             [i['xnat:mrscandata/id'] for i in exps])

    @skip_if_no_network
    def test_search_experiments(self):

        res = self._interface.array.search_experiments(project_id='nosetests3',
            experiment_type='xnat:subjectData').data
        self.assertGreaterEqual(len(res), 1)



from pyxnat import Interface
from pyxnat.core import downloadutils as du
import os.path as op
from . import skip_if_no_network

central = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

@skip_if_no_network
def test_download():
    e = central.select.project('CENTRAL_OASIS_LONG').subject('OAS2_0001').experiment('CENTRAL_E00090')
    du.download(dest_dir='/tmp', instance=e.scans(), extract=True, removeZip=True)



from pyxnat import Interface
import os.path as op
from . import skip_if_no_network
from nose import SkipTest

_modulepath = op.dirname(op.abspath(__file__))

central = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

@skip_if_no_network
def test_global_scan_listing():
    assert central.array.scans(project_id='CENTRAL_OASIS_CS',
                               experiment_type='xnat:mrSessionData',
                               scan_type='xnat:mrScanData'
                               )



import csv
import difflib
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
try:
    unicode
except NameError:
    unicode = str
from six import string_types

from lxml import etree

from .jsonutil import JsonTable, get_column, get_where, get_selection
from .errors import is_xnat_error, catch_error
from .errors import ProgrammingError, NotSupportedError
from .errors import DataError, DatabaseError
from .uriutil import check_entry

search_nsmap = {'xdat':'http://nrg.wustl.edu/security',
                'xsi':'http://www.w3.org/2001/XMLSchema-instance'}

special_ops = {'*':'%', }


def build_search_document(root_element_name, columns, criteria_set,
                          brief_description='', long_description='',
                          allowed_users=[]):
    root_node = \
        etree.Element(etree.QName(search_nsmap['xdat'], 'bundle'),
                      nsmap=search_nsmap
                      )

    root_node.set('ID', "@%s" % root_element_name)
    root_node.set('brief-description', brief_description)
    root_node.set('description', long_description)
    root_node.set('allow-diff-columns', "0")
    root_node.set('secure', "false")

    root_element_name_node = \
        etree.Element(etree.QName(search_nsmap['xdat'], 'root_element_name'),
                      nsmap=search_nsmap
                      )

    root_element_name_node.text = root_element_name

    root_node.append(root_element_name_node)

    for i, column in enumerate(columns):
        element_name, field_ID = column.split('/', 1)

        search_field_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'search_field'),
                          nsmap=search_nsmap
                          )

        element_name_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'element_name'),
                          nsmap=search_nsmap
                          )

        element_name_node.text = element_name

        field_ID_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'field_ID'),
                          nsmap=search_nsmap
                          )

        field_ID_node.text = field_ID

        sequence_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'sequence'),
                          nsmap=search_nsmap
                          )

        sequence_node.text = str(i)

        type_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'type'),
                          nsmap=search_nsmap
                          )

        type_node.text = 'string'

        header_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'header'),
                          nsmap=search_nsmap
                          )

        header_node.text = column

        search_field_node.extend([element_name_node,
                                  field_ID_node,
                                  sequence_node,
                                  type_node, header_node
                                  ])

        root_node.append(search_field_node)

    search_where_node = \
        etree.Element(etree.QName(search_nsmap['xdat'], 'search_where'),
                      nsmap=search_nsmap
                      )

    root_node.append(build_criteria_set(search_where_node, criteria_set))

    if allowed_users != []:

        allowed_users_node = \
            etree.Element(etree.QName(search_nsmap['xdat'], 'allowed_user'),
                          nsmap=search_nsmap
                          )

        for allowed_user in allowed_users:
            login_node = \
                etree.Element(etree.QName(search_nsmap['xdat'], 'login'),
                              nsmap=search_nsmap
                              )
            login_node.text = allowed_user
            allowed_users_node.append(login_node)

        root_node.append(allowed_users_node)

    return etree.tostring(root_node.getroottree())

def build_criteria_set(container_node, criteria_set):

    for criteria in criteria_set:
        if isinstance(criteria, string_types):
            container_node.set('method', criteria)

        if isinstance(criteria, (list)):
            sub_container_node = \
                etree.Element(etree.QName(search_nsmap['xdat'], 'child_set'),
                              nsmap=search_nsmap
                              )

            container_node.append(
                build_criteria_set(sub_container_node, criteria))

        if isinstance(criteria, (tuple)):
            if len(criteria) != 3:
                raise ProgrammingError('%s should be a 3-element tuple' %
                                        str(criteria)
                                        )

            constraint_node = \
                etree.Element(etree.QName(search_nsmap['xdat'], 'criteria'),
                              nsmap=search_nsmap
                              )

            constraint_node.set('override_value_formatting', '0')

            schema_field_node = \
                etree.Element(etree.QName(search_nsmap['xdat'],
                                          'schema_field'
                                          ),
                              nsmap=search_nsmap
                              )

            schema_field_node.text = criteria[0]

            comparison_type_node = \
                etree.Element(etree.QName(search_nsmap['xdat'],
                                          'comparison_type'
                                          ),
                              nsmap=search_nsmap
                              )

            comparison_type_node.text = special_ops.get(criteria[1],
                                                        criteria[1]
                                                        )

            value_node = \
                etree.Element(etree.QName(search_nsmap['xdat'], 'value'),
                              nsmap=search_nsmap
                              )

            value_node.text = criteria[2].replace('*', special_ops['*'])

            constraint_node.extend([
                    schema_field_node, comparison_type_node, value_node])

            container_node.append(constraint_node)

    return container_node

def query_from_xml(document):
    query = {}
    root = etree.fromstring(document)
    _nsmap = root.nsmap

    query['description'] = root.get('description', default="")

    query['row'] = root.xpath('xdat:root_element_name',
                              namespaces=root.nsmap)[0].text

    query['columns'] = []

    for node in root.xpath('xdat:search_field',
                           namespaces=_nsmap):

        en = node.xpath('xdat:element_name', namespaces=root.nsmap)[0].text
        fid = node.xpath('xdat:field_ID', namespaces=root.nsmap)[0].text

        query['columns'].append('%s/%s' % (en, fid))

    query['users'] = [
        node.text
        for node in root.xpath('xdat:allowed_user/xdat:login',
                               namespaces=root.nsmap
                               )
        ]

    try:
        search_where = root.xpath('xdat:search_where',
                                  namespaces=root.nsmap)[0]

        query['constraints'] = query_from_criteria_set(search_where)
    except:
        query['constraints'] = [('%s/ID' % query['row'], 'LIKE', '%'), 'AND']

    return query

def query_from_criteria_set(criteria_set):
    query = []
    query.append(criteria_set.get('method'))
    _nsmap = criteria_set.nsmap

    for criteria in criteria_set.xpath('xdat:criteria',
                                       namespaces=_nsmap):

        _f = criteria.xpath('xdat:schema_field', namespaces=_nsmap)[0]
        _o = criteria.xpath('xdat:comparison_type', namespaces=_nsmap)[0]
        _v = criteria.xpath('xdat:value', namespaces=_nsmap)[0]

        constraint = (_f.text, _o.text, _v.text)
        query.insert(0, constraint)

    for child_set in criteria_set.xpath('xdat:child_set',
                                        namespaces=_nsmap):

        query.insert(0, query_from_criteria_set(child_set))

    return query

def rpn_contraints(rpn_exp):
    left = []
    right = []
    triple = []

    for i, t in enumerate(rpn_exp.split()):
        if t in ['AND', 'OR']:
            if 'AND' in right or 'OR' in right and left == []:
                try:
                    operator = right.pop(right.index('AND'))
                except:
                    operator = right.pop(right.index('OR'))

                left = [right[0]]
                left.append(right[1:] + [t])
                left.append(operator)

                right = []

            elif right != []:
                right.append(t)

                if left != []:
                    left.append(right)
                else:
                    left = right[:]
                    right = []

            elif right == [] and left != []:
                left = [left]
                left.append(t)
                right = left[:]
                left = []
            else:
                raise ProgrammingError('in expression %s' % rpn_exp)

        else:
            triple.append(t)
            if len(triple) == 3:
                right.append(tuple(triple))
                triple = []

    return left if left != [] else right

# ---------------------------------------------------------------

class SearchManager(object):
    """ Search interface.
        Handles operations to save and get back searches on the server.

        Examples
        --------
            >>> row = 'xnat:subjectData'
            >>> columns = ['xnat:subjectData/PROJECT',
                           'xnat:subjectData/SUBJECT_ID'
                           ]
            >>> criteria = [('xnat:subjectData/SUBJECT_ID', 'LIKE', '*'),
                            'AND'
                            ]
            >>> interface.manage.search.save('mysearch', row, columns,
                                             criteria, sharing='public',
                                             description='my first search'
                                             )
    """
    def __init__(self, interface):
        self._intf = interface

    def _save_search(self, row, columns, constraints, name, desc, sharing):
        self._intf._get_entry_point()

        name = name.replace(' ', '_')
        if sharing == 'private':
            users = [self._intf._user]
        elif sharing == 'public':
            users = []
        elif isinstance(sharing, list):
            users = sharing
        else:
            raise NotSupportedError('Share mode %s not valid' % sharing)

        self._intf._exec(
            '%s/search/saved/%s?inbody=true' % (self._intf._entry, name),
            method='PUT',
            body=build_search_document(row, columns,
                                       constraints,
                                       name, desc.replace('%', '%%'),
                                       users
                                       )
            )

    def save(self, name, row, columns, constraints,
             sharing='private', description=''):
        """ Saves a query on the XNAT server.

            Parameters
            ----------
            name: string
                Name of the query displayed on the Web Interface and
                used to get back the results.
            row: string
                Datatype from `Interface.inspect.datatypes()`.
                Usually ``xnat:subjectData``
            columns: list
                List of data fields from
                `Interface.inspect.datatypes('*', '*')`
            constraints: list
                See also: `Search.where()`
            sharing: string | list
                Define by whom the query is visible.
                If sharing is a string it may be either
                ``private`` or ``public``.
                Otherwise a list of valid logins for the XNAT server
                from `Interface.users()`.

            See Also
            --------
            :func:`Search.where`
        """
        self._save_search(row, columns, constraints,
                          name, description, sharing)


    def saved(self, with_description=False):
        """ Returns the names of accessible saved search on the server.
        """
        self._intf._get_entry_point()

        jdata = self._intf._get_json(
            '%s/search/saved?format=json' % self._intf._entry)

        if with_description:
            return [(ld['brief_description'],
                     ld['description'].replace('%%', '%'))
                    for ld in get_selection(jdata, ['brief_description',
                                                    'description'
                                                    ]
                                            )
                    if not ld['brief_description'].startswith('template_')]
        else:
            return [name
                    for name in get_column(jdata, 'brief_description')
                    if not name.startswith('template_')]

    def get(self, name, out_format='results'):
        """ Returns the results of the query saved on the XNAT server or
            the query itself to know what it does.

            Parameters
            ----------
            name: string
                Name of the saved search. An exception is raised if the name
                does not exist.
            out_format: string
                Can take the following values:
                    - results to download the results of the search
                    - xml to download the XML document defining the search
                    - query to get the pyxnat representation of the search
        """
        self._intf._get_entry_point()

        jdata = self._intf._get_json(
            '%s/search/saved?format=json' % self._intf._entry)

        try:
            search_id = get_where(jdata, brief_description=name)[0]['id']
        except IndexError:
            raise DatabaseError('%s not found' % name)

        if out_format in ['xml', 'query']:
            bundle = self._intf._exec(
                '%s/search/saved/%s' % (self._intf._entry,
                                        search_id
                                        ), 'GET')

            if out_format == 'xml':
                return bundle
            else:
                return query_from_xml(bundle)


        content = self._intf._exec(
            '%s/search/saved/%s/results?format=csv' % (self._intf._entry,
                                                       search_id), 'GET')

        results = csv.reader(StringIO(content.decode('utf-8')), delimiter=',', quotechar='"')

        headers = next(results)

        return JsonTable([dict(zip(headers, res))
                          for res in results
                          ],
                         headers
                         )

    def delete(self, name):
        """ Removes the search from the server.
        """
        self._intf._get_entry_point()

        jdata = self._intf._get_json(
            '%s/search/saved?format=json' % self._intf._entry)

        try:
            search_id = get_where(jdata, brief_description=name)[0]['id']
        except IndexError:
            raise DatabaseError('%s not found' % name)

        self._intf._exec('%s/search/saved/%s' % (self._intf._entry,
                                                 search_id
                                                 ), 'DELETE')

    def save_template(self, name, row=None, columns=[],
                      constraints=[], sharing='private', description=''):
        """
            Define and save a search template. Same as the save method, but
            the values in the constraints are used as keywords for value
            replacement when using the template.

            Parameters
            ----------
            name: string
                Name under which the template is save in XNAT. A template is
                prepended to the name so that it appear clearly as a template
                on the web interface.
            row: string
                Datatype from `Interface.inspect.datatypes()`.
                Usually ``xnat:subjectData``
            columns: list
                List of data fields from
                `Interface.inspect.datatypes('*', '*')`
            constraints: list
                See also: `Search.where()`, values are keywords for the template
            sharing: string | list
                Define by whom the query is visible.
                If sharing is a string it may be either
                ``private`` or ``public``.
                Otherwise a list of valid logins for the XNAT server
                from `Interface.users()`.
        """


        def _make_template(query):
            query_template = []

            for constraint in query:
                if isinstance(constraint, tuple):
                    query_template.append((constraint[0],
                                           constraint[1],
                                           '%%(%s)s' % constraint[2])
                                          )
                elif isinstance(constraint, (unicode, str)):
                    query_template.append(constraint)
                elif isinstance(constraint, list):
                    query_template.append(_make_template(constraint))
                else:
                    raise ProgrammingError('Unrecognized token '
                                           'in query: %s' % constraint
                                           )

            return query_template

        self._save_search(
            row, columns, _make_template(constraints),
            'template_%s' % name, description, sharing
            )

    def saved_templates(self, with_description=False):
        """ Returns the names of accessible saved search templates on the server.
        """
        self._intf._get_entry_point()

        jdata = self._intf._get_json(
            '%s/search/saved?format=json' % self._intf._entry)

        if with_description:
            return [
                (ld['brief_description'].split('template_')[1],
                 ld['description'].replace('%%', '%')
                 )
                for ld in get_selection(jdata, ['brief_description',
                                                'description'
                                                ]
                                        )
                if ld['brief_description'].startswith('template_')
                ]
        else:
            return [name.split('template_')[1]
                    for name in get_column(jdata, 'brief_description')
                    if name.startswith('template_')]

    def use_template(self, name, values):
        """
            Performs a search query using a previously saved template.

            Parameters
            ----------
            name: string
                Name of the template.
            values: dict
                Values to put in the template, get the valid keys using
                the get_template method.

            Examples
            --------
            >>> interface.manage.search.use_template(name,
                          {'subject_id':'ID',
                           'age':'32'
                           })

        """
        self._intf._get_entry_point()

        bundle = self.get_template(name, True) % values

        # have to remove search_id information before re-posting it
        _query = query_from_xml(bundle)
        bundle = build_search_document(_query['row'],
                                       _query['columns'],
                                       _query['constraints']
                                       )

        content = self._intf._exec(
            "%s/search?format=csv" % self._intf._entry, 'POST', bundle)

        results = csv.reader(StringIO(content), delimiter=',', quotechar='"')
        headers = results.next()

        return JsonTable([dict(zip(headers, res))
                          for res in results
                          ],
                         headers
                         )

    def get_template(self, name, as_xml=False):
        """ Get a saved template, either as an xml document, or as a pyxnat
            representation, with the keys to be used in the template
            between the parentheses in %()s.

            Parameters
            ----------
            name: str
                Name under which the template is saved
            as_xml: boolean
                If True returns an XML document, else return a list of
                constraints. Defaults to False.
        """
        self._intf._get_entry_point()

        jdata = self._intf._get_json(
            '%s/search/saved?format=json' % self._intf._entry)

        try:
            search_id = get_where(jdata,
                                  brief_description='template_%s' % name
                                  )[0]['id']
        except IndexError:
            raise DatabaseError('%s not found' % name)

        bundle = self._intf._exec(
            '%s/search/saved/%s' % (self._intf._entry,
                                    search_id
                                    ), 'GET')

        if as_xml:
            return bundle
        else:
            _query = query_from_xml(bundle)
            return _query['row'], _query['columns'], _query['constraints'], _query['description']

    def delete_template(self, name):
        """ Deletes a search template.
        """
        self.delete('template_%s' % name)

    def eval_rpn_exp(self, rpnexp):
        return rpn_contraints(rpnexp)


class Search(object):
    """ Define constraints to make a complex search on the database.

        This :class:`Search` is available at different places throughout
        the API:

            >>> interface.select(DATA_SELECTION).where(QUERY)
            >>> interface.manage.search.save('name', TABLE_DEFINITION, QUERY)

        Examples
        --------
            >>> query = [('xnat:subjectData/SUBJECT_ID', 'LIKE', '%'),
                         ('xnat:projectData/ID', '=', 'my_project'),
                         [('xnat:subjectData/AGE', '>', '14'),
                           'AND'
                          ],
                         'OR'
                        ]
    """
    def __init__(self, row, columns, interface):
        """ Configure the result table.

            Parameters
            ----------
            row: string
                The returned table will have one line for every matching
                occurence of this type.
                e.g. xnat:subjectData
                --> table with one line per matching subject
            columns: list
                The returned table will have all the given columns.
        """
        self._row = row
        self._columns = columns
        self._intf = interface

    def where(self, constraints=None, template=None, query=None):
        """ Triggers the search.

            Parameters
            ----------
            contraints: list
                A query is an unordered list that contains
                    - 1 or more constraints
                    - 0 or more sub-queries (lists as this one)
                    - 1 comparison method between the constraints
                        ('AND' or 'OR')
                A constraint is an ordered tuple that contains
                    - 1 valid searchable_type/searchable_field
                    - 1 operator among '=', '<', '>', '<=', '>=', 'LIKE'

            Returns
            -------
            results: JsonTable object
                An table-like object containing the results. It is
                basically a list of dictionaries that has additional
                helper methods.
        """
        self._intf._get_entry_point()

        if isinstance(constraints, (str, unicode)):
            constraints = rpn_contraints(constraints)
        elif isinstance(template, (tuple)):
            tmp_bundle = self._intf.manage.search.get_template(
                template[0], True)

            tmp_bundle = tmp_bundle % template[1]
            constraints = query_from_xml(tmp_bundle)['constraints']
        elif isinstance(query, (str, unicode)):
            tmp_bundle = self._intf.manage.search.get(query, 'xml')
            constraints = query_from_xml(tmp_bundle)['constraints']
        elif isinstance(constraints, list):
            pass
        else:
            raise ProgrammingError('One of contraints, template and query'
                                   'parameters must be correctly set.')

        bundle = build_search_document(self._row, self._columns, constraints)

        content = self._intf._exec(
            "%s/search?format=csv" % self._intf._entry, 'POST', bundle)

        if is_xnat_error(content):
            catch_error(content)

        results = csv.reader(StringIO(content.decode('utf-8')), delimiter=',', quotechar='"')
        headers = next(results)

        headers_of_interest = []

        for column in self._columns:
            try:
                headers_of_interest.append(
                    difflib.get_close_matches(
                        column.split(self._row + '/')[0].lower() \
                            or column.split(self._row + '/')[1].lower(),
                        headers)[0]
                    )
            except IndexError:
                headers_of_interest.append('unknown')

        if len(self._columns) != len(headers_of_interest):
            raise DataError('unvalid response headers')

        return JsonTable([dict(zip(headers, res)) for res in results],
                         headers_of_interest).select(headers_of_interest)

    def all(self):
        return self.where([(self._row + '/ID', 'LIKE', '%'), 'AND'])



import os.path as op
import tempfile

from pyxnat import jsonutil

_csv_example = op.join(op.dirname(op.abspath(__file__)),
                            'results.csv')
_list_of_dirs = jsonutil.csv_to_json(open(_csv_example, 'r').read())

jtable = jsonutil.JsonTable(_list_of_dirs)

# coverage_tests
def test_dump_methods():
    jtable.__repr__()

    jtable.dumps_csv()
    jtable.dumps_json()

    jtable.dump_csv(tempfile.mkstemp()[1])
    jtable.dump_json(tempfile.mkstemp()[1])

    jtable.where(psytool_audit_parentdata_audit9='0',
                 psytool_audit_parentdata_audit2='0',
                 psytool_audit_parentdata_audit1='2',
                 psytool_audit_parentdata_audit10='2')

    jtable.where('5154')
    jtable.where_not('5154')

    assert True

# unit_tests
def test_get_item_str_value():
    jtable.items()
    assert jtable['subject_label'] == jtable.get('subject_label')

def test_get_item_int_value():
    assert jtable.data[0] == jtable[0].data[0]

def test_get_item_list_value():
    assert set(jtable[['projects', 'subjectid']].headers()) == set(['projects', 'subjectid'])

def test_join():
    index = list(jtable.headers()).index('subjectid')
    a = jtable.select(list(jtable.headers())[:index+1])
    b = jtable.select(list(jtable.headers())[index:])
    c = a.join('subjectid', b)
    assert len(list(jtable.headers())) == len(list(c.headers()))



import os.path as op
from pyxnat import Interface
from . import skip_if_no_network
from nose import SkipTest

central = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

@skip_if_no_network
def test_001_xpath_checkout():
    central.xpath.checkout(subjects=['OAS1_0001', 'OAS1_0002'])
    assert 'OAS1_0001' in central.xpath.subjects() and \
        'OAS1_0002' in central.xpath.subjects()

@skip_if_no_network
def test_elements():
    assert 'fs:region' in central.xpath.elements()

@skip_if_no_network
def test_keys():
    assert 'ID' in central.xpath.keys()

@skip_if_no_network
def test_values():
    assert 'OAS1_0002' in central.xpath.values('ID')

@skip_if_no_network
def test_element_attrs():
    assert isinstance(central.xpath.element_attrs('fs:region'), list)

    assert set(['SegId', 'hemisphere', 'name']).issubset(
        central.xpath.element_keys('fs:region'))

    assert 'Left-Putamen' in \
        central.xpath.element_values('fs:region', 'name')



import difflib
import six
if six.PY3:
    from urllib.parse import quote
elif six.PY2:
    from urllib import quote

from .jsonutil import JsonTable
from .uriutil import uri_parent
from .schema import datatype_attributes


class EAttrs(object):
    """ Accessor class to resource fields.

        Help to retrieve the attributes paths relevant to this element::

            >>> subject.attrs()
            ['xnat:subjectData/sharing',
             'xnat:subjectData/sharing/share',
             'xnat:subjectData/resources',
             ...
             'xnat:subjectData/experiments/experiment'
             ]

         All paths are not valid but they give an indication of what
         is available. To retrieve the paths, the corresponding
         schemas must be downloaded first through the schema
         management interface in order to be parsed::
         
             >>> interface.manage.schemas.add('xnat.xsd')
             >>> interface.manage.schemas.add('myschema/myschema.xsd')
    """
    def __init__(self, eobj):
        """
            Parameters
            ----------
            eobj:
                :class:`EObject` Object
        """
        self._eobj = eobj
        self._intf = eobj._intf
        self._datatype = None
        self._id = None

    def __call__(self):
        """ List the attributes paths relevant to this element.
        """
        paths = []
        for root in self._intf.manage.schemas._trees.values():
            paths.extend(datatype_attributes(root, self._get_datatype()))
        return paths

    def _get_datatype(self):
        if self._datatype is None:
            self._datatype = self._eobj.datatype()

        return self._datatype

    def _get_id(self):
        if self._id is None:
            self._id = self._eobj.id()

        return self._id

    def set(self, path, value, **kwargs):
        """ Set an attribute.

            Parameters
            ----------
            path: string
                The xpath of the attribute relative to the element.
            value: string
                The attribute's value. Note that the python type is
                always a string but the content of the value must
                match what is defined in the schema.
                e.g. an element defined as a float in the schema
                must be given a string containing a number, a
                valid date must follow the ISO 8601 which is the
                standard representation for dates and times
                established by the W3C.
        """
        dt = self._get_datatype()
        if dt is None:
            dt = ''
        put_uri = self._eobj._uri + '?xsiType=%s&%s=%s' % (quote(dt)
                                                         , quote(path)
                                                         , quote(value)
                                              )

        self._intf._exec(put_uri, 'PUT', **kwargs)

    def mset(self, dict_attrs, **kwargs):
        """ Set multiple attributes at once.

            It is more efficient to use this method instead of
            multiple times the `set()` method when setting more than
            one attribute because only a single HTTP call is issued to
            the server.

            Parameters
            ----------
            dict_attrs: dict
                The dict of key values to set. It follows the same
                principles as the single `set()` method.
        """

        query_str = '?xsiType=%s' % (quote(self._get_datatype())) + ''.join(['&%s=%s' % (quote(path),
                                               quote(val)
                                               )
                                    for path, val in dict_attrs.items()
                                    ]
                                   )

        put_uri = self._eobj._uri + query_str

        self._intf._exec(put_uri, 'PUT', **kwargs)

    def get(self, path):
        """ Get an attribute value.

            .. note::
                The value is always returned in a Python string. It must
                be explicitly casted or transformed if needed.

            Parameters
            ----------
            path: string
                The xpath of the attribute relative to the element.

            Returns
            -------
            A string containing the value.
        """
        query_str = '?columns=ID,%s' % path

        get_uri = uri_parent(self._eobj._uri) + query_str
        jdata = JsonTable(self._intf._get_json(get_uri)
                          ).where(ID=self._get_id())

        # unfortunately the return headers do not always have the
        # expected name

        header = difflib.get_close_matches(path.split('/')[-1],
                                           jdata.headers()
                                           )

        if header == []:
            header = difflib.get_close_matches(path, jdata.headers())[0]
        else:
            header = header[0]

        replaceSlashS = lambda x : x.replace('\s', ' ')
        if type(jdata.get(header)) == list:
            return map(replaceSlashS, jdata.get(header))
        else:
            return jdata.get(header).replace('\s', ' ')

    def mget(self, paths):
        """ Set multiple attributes at once.

            It is more efficient to use this method instead of
            multiple times the `get()` method when getting more than
            one attribute because only a single HTTP call is issued to
            the server.

            Parameters
            ----------
            paths: list
                List of attributes' paths.

            Returns
            -------
            list: ordered list of values (in the order of the
            requested paths)
        """

        query_str = '?columns=ID,%s' % ','.join(paths)
        get_uri = uri_parent(self._eobj._uri) + query_str

        jdata = JsonTable(self._intf._get_json(get_uri)
                          ).where(ID=self._get_id())

        results = []

        # unfortunately the return headers do not always have the
        # expected name

        for path in paths:
            header = difflib.get_close_matches(path.split('/')[-1],
                                               jdata.headers())

            if header == []:
                header = difflib.get_close_matches(path, jdata.headers())[0]
            else:
                header = header[0]
            results.append(jdata.get(header).replace('\s', ' '))

        return results



from functools import wraps
from nose import SkipTest
from nose.plugins.attrib import attr
import os


def skip_if_no_network(func=None):
    """Skip test completely in NONETWORK settings
    If not used as a decorator, and just a function, could be used at the module level
    """

    def check_and_raise():
        if os.environ.get('PYXNAT_SKIP_NETWORK_TESTS'):
            raise SkipTest("Skipping since no network settings")

    if func:
        @wraps(func)
        @attr('skip_if_no_network')
        def newfunc(*args, **kwargs):
            check_and_raise()
            return func(*args, **kwargs)
        return newfunc
    else:
        check_and_raise()



import urllib

from lxml import etree

from .search import SearchManager
from .users import Users
from .resources import Project
from .tags import Tags
from .jsonutil import JsonTable


class GlobalManager(object):
    """ Mainly a container class to provide a clean interface for all
        management classes.
    """

    def __init__(self, interface):
        self._intf = interface

        self.search = SearchManager(self._intf)
        self.users = Users(self._intf)
        self.tags = Tags(self._intf)
        self.schemas = SchemaManager(self._intf)
        self.prearchive = PreArchive(self._intf)

    def register_callback(self, func):
        """ Defines a callback to execute when collections of resources are
            accessed.

            Parameters
            ----------
            func: callable
                A callable that takes the current collection object as first
                argument and the current element object as second argument.

            Examples
            --------
            >>> def notify(cobj, eobj):
            >>>    print eobj._uri
            >>> interface.manage.register_callback(notify)
        """
        self._intf._callback = func

    def unregister_callback(self):
        """ Unregisters the callback.
        """
        self._intf._callback = None

    def project(self, project_id):
        """ Returns a project manager.
        """
        return ProjectManager(project_id, self._intf)


class ProjectManager(object):
    """ Management interface for projects.

        This functionalities are also available through `Project` objects.
    """

    def __init__(self, project_id, interface):

        self._intf = interface
        self._intf._get_entry_point()

        project = Project('%s/projects/%s' % (self._intf._entry,
                                              project_id
                                              ),
                          self._intf
                          )

        self.prearchive_code = project.prearchive_code
        self.set_prearchive_code = project.set_prearchive_code
        self.quarantine_code = project.quarantine_code
        self.set_quarantine_code = project.set_quarantine_code
        self.current_arc = project.current_arc
        self.set_subfolder_in_current_arc = \
            project.set_subfolder_in_current_arc
        self.accessibility = project.accessibility
        self.users = project.users
        self.owners = project.owners
        self.members = project.members
        self.collaborators = project.collaborators
        self.user_role = project.user_role
        self.add_user = project.add_user
        self.remove_user = project.remove_user


class SchemaManager(object):
    """ Management interface for XNAT schemas.

        The aim is to provide a minimal set of functionalities to parse
        and look at XNAT schemas.
    """

    def __init__(self, interface):
        self._intf = interface
        self._trees = {}

    def __call__(self):
        return self._trees.keys()

    def add(self, url):
        """ Loads an additional schema.

            Parameters
            ----------
            url: str
                url of the schema relative to the server.
                    e.g. for
                    http://central.xnat.org/xapi/schemas/xnat give
                    ``/xapi/schemas/xnat``

        """

        #if not re.match('/?schemas/.*/.*\.xsd', url):
        #    if not 'schemas' in url and re.match('/?\w+/\w+[.]xsd', url):
        #        url = join_uri('/schemas', url)
        #
        #    elif not re.match('^[^/].xsd', url):
        #        url = '/schemas/%s/%s'%(url.split('.xsd')[0], url)
        #    else:
        #        raise NotImplementedError

        self._trees[url.split('/')[-1]] = \
            etree.fromstring(self._intf._exec(url))

    def remove(self, name):
        """ Removes a schema.
        """
        if name in self._trees.keys():
            del self._trees[name]

"""
   Fields Of the Prearchive
   ------------------------
   Each session in the prearchive
    has the following fields:
       "project" - The name of the project. "Unassigned" if the session is unassigned.
       "timestamp" - The time (down to millisecond) that this session was received
                     by XNAT. "20110603_124835868" for example.
       "lastmod" - The time this session as last modified. Moving, resetting etc.
                   updates this time
       "uploaded" - The time this session was uploaded to XNAT. Usually the same as
                    "timestamp"
       "scan_date" - The date this session was scanned.
       "scan_time" - The time this session was scanned.
       "subject" - The name of the subject
       "folderName" - The id of this session. Corresponds to XNAT's session id,
       "name" - The name of this session. Corresponds to XNAT's session label.
       "tag" - This session's unique DICOM identifier. Usually the SOP instance UID.
       "status" - The current status of this session
       "url" - The unique uri of this session.
       "autoarchive" - Whether this session should be auto-archived.
    For more in-depth information about the prearchive see:
    http://docs.xnat.org/Using+the+Prearchive#XNAT%201.5:%20Managing%20Data%20with%20the%20Prearchive-Operations-Session%20Status

    Uniquely identifying a session
    ------------------------------
    Every session in the prearchive uniquely is identified by "project",
    "timestamp" and "folderName".
    (13/7/2011) - Each session could have been uniquely identified by the "url"
                  field or the "tag" field. These are arguably more elegant
                  identifiers but for now the "project", "timestamp", "folderName"
                  triple is used.

"""
class PreArchive(object):
    def __init__(self, interface):
        self._intf = interface

    """
    Retrieve the status of a session
    Parameters
    ----------
       triple - A list containing the project, timestamp and session id, in that order.
    """
    def status(self, triple):
        return JsonTable(
            self._intf._get_json('/data/prearchive/projects')
            ).where(
            project=triple[0], timestamp=triple[1], folderName=triple[2]
            ).get('status')

    """
    Retrieve the contents of the prearchive.
    """
    def get(self):
        return JsonTable(self._intf._get_json('/data/prearchive/projects'),
                         ['project', 'timestamp', 'folderName']
                         ).select(['project', 'timestamp', 'folderName']
                                  ).as_list()[1:]

    """
    Retrieve the scans of a give session triple
    Parameters
    ----------
       triple - A list containing the project, timestamp and session id, in that order.
    """
    def get_scans(self, triple):
        return JsonTable(self._intf._get_json(
                '/data/prearchive/projects/%s/scans' \
                    % '/'.join(triple)
                )).get('ID')

    """
    Retrieve the resource of a session triple
    Parameters
    ----------
       triple - A list containing the project, timestamp and session id, in that order.
       scan_id - id of the scan
    """
    def get_resources(self, triple, scan_id):
        return JsonTable(self._intf._get_json(
                '/data/prearchive/projects/%s'
                '/scans/%s/resources' % ('/'.join(triple), scan_id)
                )).get('label')

    """
    Retrieve a list of files in a given session triple
    Parameters
    ----------
       triple - A list containing the project, timestamp and session id, in that order.
       scan_id - id of the scan
       resource_id - id of the resource
    """
    def get_files(self, triple, scan_id, resource_id):
        return JsonTable(self._intf._get_json(
                '/data/prearchive/projects/%s'
                '/scans/%s/resources/%s/files' % ('/'.join(triple),
                                                  scan_id,
                                                  resource_id)
                )).get('Name')

    """
    Move multiple sessions to a project in the prearchive asynchronously.
    If only one session is it is done now.

    This does *not* archive a session.

    Parameters
    ----------
       uris - a list of session uris
       new_project - The name of the project to which to move the sessions.
    """
    def move (self, uris, new_project):
        add_src = lambda u: urllib.urlencode({'src':u})

        async_ = len(uris) > 1 and 'true' or 'false'
        print(async_)

        post_body = '&'.join ((map(add_src,uris))
                            + [urllib.urlencode({'newProject':new_project})]
                            + [urllib.urlencode({'async':async_})])

        request_uri = '/data/services/prearchive/move?format=csv'
        return self._intf._exec(request_uri ,'POST', post_body,

                                {'content-type':'application/x-www-form-urlencoded'})
    """
    Reinspect the file on the filesystem on the XNAT server and recreate the
    parameters of the file. Essentially a refresh of the file.

    Be warned that if this session has been scheduled for an operation, that
    operation is cancelled.
    Parameters
    ----------
       uris - a list of session uris
       new_project - The name of the project to which to move the sessions.
    """
    def reset(self, triple):
        post_body = "action=build"
        request_uri = '/data/prearchive/projects/%s?format=single' \
                    % '/'.join(triple)
        return self._intf._exec(request_uri ,'POST', post_body,
                                {'content-type':'application/x-www-form-urlencoded'})
    """
    Delete  a session from the prearchive
    Parameters
    ----------
       uri - The uri of the session to delete
    """
    def delete (self, uri):
        post_body = "src=" + uri + "&" + "async=false"
        request_uri = "/data/services/prearchive/delete?format=csv"
        return self._intf._exec(request_uri ,'POST', post_body,
                                {'content-type':'application/x-www-form-urlencoded'})
    """
    Get the uri of the given session.
    Parameters
    ----------
       triple - A list containing the project, timestamp and session id, in that order.
    """
    def get_uri(self, triple):
        return JsonTable(
            self._intf._get_json('/data/prearchive/projects')
            ).where(
            project=triple[0], timestamp=triple[1], folderName=triple[2]
            ).get('url')



import os.path as op
from pyxnat import Interface

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')

central = Interface(config=fp)

def test_ashs_volumes():
    r = central.select.experiment('CENTRAL04_E00637').resource('ASHS')
    hv = r.volumes()
    v = hv.query('region=="CA1" & side=="left"')['volume'].tolist()[0]
    assert(v == 1585.838)



from uuid import uuid1
from pyxnat import manage
from pyxnat import Interface
import os.path as op

x = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

def test_project_manager():
    pm = manage.ProjectManager('nosetests', x)



from pyxnat import Interface
import os.path as op
from . import skip_if_no_network

central = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

@skip_if_no_network
def test_prearchive_get():
    from pyxnat.core import manage
    pa = manage.PreArchive(central)
    pc = pa.get()

@skip_if_no_network
def test_prearchive_status():
    triple = ['SAFMD', '20170602_161757472', '_4']
    from pyxnat.core import manage
    pa = manage.PreArchive(central)
    assert(pa.status(triple) == 'READY')
    pa.get_uri(triple)



from pyxnat import Interface
import os.path as op
from . import skip_if_no_network

central = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))

def test_inspector_structure():
    from pyxnat.core import Inspector
    i = Inspector(central)
    i.set_autolearn()
    print(i.datatypes())
    s = i.structure()



import os
import time
import getpass
import json
import requests

try:
    import socks
except ImportError:
    socks = None

import six
if six.PY2:
    from urlparse import urlparse
    input = raw_input
elif six.PY3:
    from urllib.parse import urlparse

from .select import Select
from .help import Inspector, GraphData, PaintGraph, _DRAW_GRAPHS
from .manage import GlobalManager
from .uriutil import join_uri, file_path, uri_last
from .jsonutil import csv_to_json
from .errors import is_xnat_error
from .errors import catch_error
from .array import ArrayData
from .xpath_store import XpathStore
from . import xpass

DEBUG = False
STUBBORN = False


# main entry point
class Interface(object):
    """ Main entry point to access an XNAT server.

        >>> central = Interface(server='http://central.xnat.org:8080',
                                user='login',
                                password='pwd')

        Or with config file:

        >>> central = Interface(config='/home/me/.xnat.cfg')

        Or for interactive use:

        >>> central = Interface('http://central.xnat.org')

        .. note::
            The interactive mode is triggered whenever an argument (between
            server, user or password) is missing. In interactive mode pyxnat
            will check that connection settings are valid.

        .. note::
            Proxy support requires the socks module be installed. This can be
            installed via pip::

            `pip install SocksiPy-branch`

        Or anonymously (unauthenticated):

        >>> central = Interface('http://central.xnat.org', anonymous=True)
    """

    def __init__(self, server=None, user=None, password=None, config=None,
                 anonymous=False, proxy=None, verify=None):
        """
            Parameters
            ----------
            server: string | None
                The server's full URL (including port and XNAT instance name
                if necessary) e.g. http://central.xnat.org,
                http://localhost:8080/xnat_db
                If None the user will be prompted for it.
            user: string | None
                A valid login registered through the XNAT web interface.
                If None the user will be prompted for it.
            password: string | None
                The user's password.
                If None the user will be prompted for it.

            config: string
                Reads a config file in json to get the connection parameters.
                If a config file is specified, it will be used regardless of
                other parameters that might have been given.
            anonymous: boolean
                Indicates an unauthenticated connection.  If True, user
                and password are ignored and a session is started with
                no credentials.
            proxy: string | None
                Indicates the full URL for an HTTP proxy server to be used for
                transactions with the specified XNAT server. If you need to
                specify a username and password for proxy access, prepend them
                to the hostname:
                http://user:pass@hostname:port

            verify: True, False, or path to file containing certificate for your site
              Added to the requests Session, as documented here:
              http://docs.python-requests.org/en/latest/user/advanced/#ssl-cert-verification
              Simplifies handling self-certified sites, or sites where there is an issue
              with certification

        """

        self._interactive = False
        self._anonymous = anonymous
        self._verify = verify

        if self._anonymous:

            if server is None:
                self._server = input('Server: ')
                self._interactive = True
            else:
                self._server = server
                self._interactive = False

            self.__set_proxy(proxy)

            self._user = None
            self._pwd = None

        else:

            if not all([server, user, password]) and not config:
                self._interactive = True

            if all(arg is None
                    for arg in [server, user, password, config]) \
                    and os.path.exists(xpass.path()):

                connection_args = xpass.read_xnat_pass(xpass.path())

                if connection_args is None:
                    raise Exception('XNAT configuration file not found '
                                    'or formatted incorrectly.')

                self._server = connection_args['host']
                self._user = connection_args['u']
                self._pwd = connection_args['p']


                if 'proxy' in connection_args:
                    self.__set_proxy(connection_args['proxy'])
                else:
                    self.__set_proxy(None)

            elif config is not None:
                self.load_config(config)

            else:
                if server is None:
                    self._server = input('Server: ')
                else:
                    self._server = server

                if user is None:
                    user = input('User: ')

                if password is None:
                    password = getpass.getpass()

                self._user = user
                self._pwd = password

                self.__set_proxy(proxy)

        self._callback = None

        self._struct = {}
        self._entry = None
        self._jsession = None #'authentication_by_credentials'
        self._connect_extras = {}
        self._connect()

        self.inspect = Inspector(self)
        self.select = Select(self)
        self.array = ArrayData(self)
        self.manage = GlobalManager(self)
        self.xpath = XpathStore(self)

        if _DRAW_GRAPHS:
            self._get_graph = GraphData(self)
            self.draw = PaintGraph(self)

        if self._interactive:
            self._get_entry_point()

        self.inspect()

    def __getstate__(self):
        return {
            '_server': self._server,
            '_user': self._user,
            '_pwd': self._pwd,
            '_anonymous': self._anonymous,
        }

    def __setstate__(self, dictionary):
        self.__dict__ = dictionary
        if self._anonymous:
            self.__init__(self._server, anonymous=True)
        else:
            self.__init__(self._server, self._user, self._pwd)

    def __set_proxy(self, proxy=None):
        if proxy is None:
            proxy = os.environ.get("http_proxy")
        if proxy is None:
            self._proxy_url = None
        else:
            self._proxy_url = urlparse(proxy)

    def _get_entry_point(self):
        if self._entry is None:
            # /REST for XNAT 1.4, /data if >=1.5
            self._entry = '/REST'
            try:
                ans = self._exec('/data/JSESSION', force_preemptive_auth=True)
                self._jsession = 'JSESSIONID=' + str(ans)
                self._entry = '/data'

                if is_xnat_error(self._jsession):
                    catch_error(self._jsession)
            except Exception as e:
                if not '/data/JSESSION' in str(e):
                    raise e

        return self._entry

    def _connect(self, **kwargs):
        """ Sets up the connection with the XNAT server.

            Parameters
            ----------
            kwargs: dict
                Can be used to pass additional arguments to
                the Http constructor. See the httplib2 documentation
                for details. http://code.google.com/p/httplib2/
        """

        if kwargs != {}:
            self._connect_extras = kwargs
        else:
            kwargs = self._connect_extras

        self._http = requests.Session()

        # requests verify defaults to True, but can be set from environment variables
        # Leave as-is unless user has explicitly overridden it
        if self._verify is not None:
          self._http.verify = self._verify

        if not self._anonymous:
            self._http.auth = (self._user, self._pwd)

        if self._proxy_url:
            self._http.proxies = {'http': self._proxy_url.geturl()}


        # Turns out this doesn't work any more: XNAT doesn't do the 401 response that forces
        # httplib2 to re-submit the request with credentials. See where the Authorization header
        # is added manually in the _exec function.
        # if not self._anonymous:
        #    self._http.add_credentials(self._user, self._pwd)

    def _exec(self, uri, method='GET', body=None, headers=None, force_preemptive_auth=False, **kwargs):
        """ A wrapper around a simple httplib2.request call that:
                - avoids repeating the server url in the request
                - deals with custom caching mechanisms :: Depricated
                - manages a user session with cookies
                - catches and broadcast specific XNAT errors

            Parameters
            ----------
            uri: string
                URI of the resource to be accessed. e.g. /REST/projects
            method: GET | PUT | POST | DELETE | HEAD
                HTTP method.
            body: string | dict
                HTTP message body
            headers: dict
                Additional headers for the HTTP request.
            force_preemptive_auth: boolean
                .. note:: Depricated as of 1.0.0.0
                Indicates whether the request should include an Authorization header with basic auth credentials.
            **kwargs: dictionary
                Additional parameters to pass directly to the Requests HTTP call.

            HTTP:GET
            ----------
                When calling with GET as method, the body parameter can be a key:value dictionary containing
                request parameters or a string of parameters. They will be url encoded and appended to the url.

            HTTP:POST
            ----------
                When calling with POST as method, the body parameter can be a key:value dictionary containing
                request parameters they will be url encoded and appended to the url.

        """

        if headers is None:
            headers = {}

        self._get_entry_point()

        uri = join_uri(self._server, uri)

        if DEBUG:
            print(uri)

        response = None

        def request(method, uri, headers, body, kwargs):
            if method == 'PUT':
                response = self._http.put(uri, headers=headers, data=body,
                    **kwargs)
            elif method == 'GET':
                response = self._http.get(uri, headers=headers, params=body,
                    **kwargs)
            elif method == 'POST':
                response = self._http.post(uri, headers=headers, data=body,
                    **kwargs)
            elif method == 'DELETE':
                response = self._http.delete(uri, headers=headers, data=body,
                    **kwargs)
            elif method == 'HEAD':
                response = self._http.head(uri, headers=headers, data=body,
                    **kwargs)
            else:
                print('Unsupported HTTP method (%s)'%method)
                return
            return response

        response = request(method, uri, headers, body, kwargs)
        if response is None:
            return

        # Dirty trick to help Travis CI tests on CENTRAL: consider fixing it
        if STUBBORN:
            if (response is not None and response.status_code == 500):
                print('Retrying request... %s'%uri)
                response = request(method, uri, headers, body, kwargs)

        if (response is not None and not response.ok) or is_xnat_error(response.content):
            if DEBUG:
                print(response.content)
                print(response.status_code)

            catch_error(response.content, '''pyxnat._exec failure:
                    URI: {response.url}
                    status code: {response.status_code}
                    headers: {response.headers}
                    content: {response.content}
                '''.format(response=response))

        return response.content

    def _get_json(self, uri):
        """ Specific Interface._exec method to retrieve data.
            It forces the data format to csv and then puts it back to a
            json-like format.

            Parameters
            ----------
            uri: string
                URI of the resource to be accessed. e.g. /REST/projects

            Returns
            -------
            List of dicts containing the results
        """
        if 'format=json' in uri:
            uri = uri.replace('format=json', 'format=csv')
        else:
            if '?' in uri:
                uri += '&format=csv'
            else:
                uri += '?format=csv'

        content = self._exec(uri, 'GET')

        if is_xnat_error(content):
            catch_error(content)

        json_content = csv_to_json(content)

        # add the (relative) path field for files
        base_uri = uri.split('?')[0]
        if uri_last(base_uri) == 'files':
            for element in json_content:
                element['path'] = file_path(element['URI'])
        return json_content

    def _get_head(self, uri):
        if DEBUG:
            print('GET HEAD')

        response = self._http.head('{server}{uri}'.format(server=self._server, uri=uri))

        if not response.ok:
            time.sleep(1)
            self._http.head('{server}{uri}'.format(server=self._server, uri=uri))

        return response.headers

    def save_config(self, location):
        """ Saves current configuration - including password - in a file.

            .. warning::
                Since the password is saved as well, make sure the file
                is saved at a safe location with appropriate permissions.

            .. note::
                This method raises NotImplementedError for an anonymous
                interface.

            Parameters
            ----------
            location: string
                Destination config file.
        """
        if self._anonymous:
            raise NotImplementedError(
                'no save_config() for anonymous interfaces')

        if not os.path.exists(os.path.dirname(location)):
            os.makedirs(os.path.dirname(location))

        config = {'server': self._server,
                  'user': self._user,
                  'password': self._pwd,
                  }
        if self._verify:
            config['verify'] = self._verify

        if self._proxy_url:
            config['proxy'] = self._proxy_url.geturl()

        with open(location, 'w') as fp:
            json.dump(config, fp)


    def load_config(self, location):
        """ Loads a configuration file and replaces current connection
            parameters.

            .. note::
                This method raises NotImplementedError for an anonymous
                interface.

            Parameters
            ----------
            location: string
                Configuration file path.
        """
        if self._anonymous:
            raise NotImplementedError(
                'no load_config() for anonymous interfaces')

        if os.path.exists(location):
            with open(location, 'rb') as fp:
                config = json.load(fp)


            self._server = str(config['server'])
            self._user = str(config['user'])
            self._pwd = str(config['password'])

            self._verify = bool(config.get('verify', True))
            if 'proxy' in config:
                self.__set_proxy(str(config['proxy']))
            else:
                self.__set_proxy(None)

        else:
            raise Exception('Configuration file does not exist.')

    def version(self):
        """
            Get version of the currently running XNAT instance.
        """
        from pyxnat.core.errors import DatabaseError
        try:
            return self._exec('/data/version')
        except DatabaseError:
            return json.loads(self._exec('/xapi/siteConfig/buildInfo').decode())

    def set_logging(self, level=0):
        pass

    def disconnect(self):
        """
            Tell XNAT to disconnect this session
        """
        self._exec('/data/JSESSION', method='DELETE')
        self._http.close()


    def get(self, uri, **kwargs):
        '''
        Wrapper around requests.get()
        returns rquests.response object
        '''
        uri = join_uri(self._server, uri)
        response = self._http.get(uri, **kwargs)
        return response

    def put(self, uri, **kwargs):
        '''
        Wrapper around requests.put()
        returns rquests.response object
        '''
        uri = join_uri(self._server, uri)
        response = self._http.put(uri, **kwargs)
        return response

    def post(self, uri, **kwargs):
        '''
        Wrapper around requests.post()
        returns rquests.response object
        '''
        uri = join_uri(self._server, uri)
        response = self._http.post(uri, **kwargs)
        return response

    def delete(self, uri, **kwargs):
        '''
        Wrapper around requests.delete()
        returns rquests.response object
        '''
        uri = join_uri(self._server, uri)
        response = self._http.delete(uri, **kwargs)
        return response

    def head(self, uri, **kwargs):
        '''
        Wrapper around requests.head()
        returns rquests.response object
        '''
        uri = join_uri(self._server, uri)
        response = self._http.head(uri, **kwargs)
        return response

    def __enter__(self):

        return self

    def close_jsession(self):
        '''
        Closes the session with XNAT server and consumes the JSESSIONID token.
        '''
        uri = '/data/JSESSION'
        response = self.delete(uri)
        response.raise_for_status()
        self._jsession = None

    def __exit__(self, type, value, traceback):

        if self._jsession:
            self.close_jsession()
        self._http.close()



import os
import os.path as op
from uuid import uuid1
import time
from . import skip_if_no_network
from pyxnat import Interface

_modulepath = op.dirname(op.abspath(__file__))

central = Interface(config=op.join(op.dirname(op.abspath(__file__)), 'central.cfg'))
from pyxnat.core import interfaces
interfaces.STUBBORN = True

sid = uuid1().hex
eid = uuid1().hex

subject = central.select.project('nosetests3').subject(sid)
experiment = subject.experiment(eid)

@skip_if_no_network
def test_01_fancy_resource_create():

    field_data = {'experiment': 'xnat:mrSessionData',
                  'ID': 'TEST_%s' % eid,
                  'xnat:mrSessionData/age': '42',
                  'xnat:subjectData/investigator/lastname': 'doe',
                  'xnat:subjectData/investigator/firstname': 'john',
                  'xnat:subjectData/ID': 'TEST_%s' % sid,
                  }

    experiment.create(**field_data)
    assert subject.exists()
    assert experiment.exists()

    globals()['subject'] = experiment.parent()
    globals()['experiment'] = experiment

@skip_if_no_network
def test_02_attr_get():
        assert experiment.attrs.get('xnat:mrSessionData/age') == '42.0'

@skip_if_no_network
def test_03_attr_mget():
    time.sleep(5)
    fields = ['xnat:subjectData/investigator/firstname',
              'xnat:subjectData/investigator/lastname'
              ]

    assert subject.attrs.mget(fields) == ['john', 'doe']

@skip_if_no_network
def test_04_attr_set():
    experiment.attrs.set('xnat:mrSessionData/age', '26')
    assert experiment.attrs.get('xnat:mrSessionData/age') == '26.0'

# def test_attr_mset():
#     subject = central.select.project('nosetests').subject(sid)
#     time.sleep(5)
#     field_data = {'xnat:subjectData/investigator/firstname':'angus',
#                   'xnat:subjectData/investigator/lastname':'young',
#                   }
#
#     subject.attrs.mset(field_data)
#     returned = subject.attrs.mget(field_data.keys())
#
#     assert set(returned) == \
#         set(field_data.values()), '''set: %s returned: %s ''' %(field_data.values(), returned)

@skip_if_no_network
def test_05_cleanup():

    subject.delete()
    assert not subject.exists()

@skip_if_no_network
def test_06_list_project_attrs():

    project_attributes = ['xnat:projectData/name',
                          'xnat:projectData/type',
                          'xnat:projectData/description',
                          'xnat:projectData/keywords',
                          'xnat:projectData/aliases',
                          'xnat:projectData/aliases/alias',
                          'xnat:projectData/aliases/alias/None',
                          'xnat:projectData/publications',
                          'xnat:projectData/publications/publication',
                          'xnat:projectData/resources',
                          'xnat:projectData/resources/resource',
                          'xnat:projectData/studyProtocol',
                          'xnat:projectData/PI',
                          'xnat:projectData/investigators',
                          'xnat:projectData/investigators/investigator',
                          'xnat:projectData/fields',
                          'xnat:projectData/fields/field',
                          'xnat:projectData/fields/field/None']

    p = central.select.project('nosetests')
    assert(p.attrs() == [])

    central.manage.schemas.add('xapi/schemas/xnat')
    p = central.select.project('nosetests')
    assert (p.attrs() == project_attributes)



XNAT_RESOURCE_NAMES = ['BBRC_VALIDATOR']

def tests(self, name, key=None):
    import json
    j = [e for e in list(self.files('{}*.json'.format(name)))][0]
    j = json.loads(self._intf.get(j._uri).text)
    if key is None:
        return j
    else:
        return j[key]

def pdf(self, name):
    import logging as log
    files = list(self.files())
    pdf = {each._uri.split('/')[-1]:each for each in files \
        if name in each._uri.split('/')[-1] and \
        each._uri.split('/')[-1].endswith('.pdf')}

    if len(pdf.items()) == 0:
        raise Exception('No %s found'%name)

    keys = list(pdf.keys())
    if len(keys) != 1:
        raise Exception('Multiple matching reports found (%s)'%keys)

    f = pdf[list(pdf.keys())[0]]
    return f

def download_snapshot(self, name, fp):
    import os.path as op
    import shutil
    import os
    import os.path as op
    import tempfile
    import logging as log
    from glob import glob

    def extract_file(fp):
        dn = op.dirname(fp)

        with open(fp, "rb") as file:
            pdf = file.read()

        startmark = b"\xff\xd8"
        startfix = 0
        endmark = b"\xff\xd9"
        endfix = 2
        i = 0

        njpg = 0
        jpeg = []
        while True:
            istream = pdf.find(b"stream", i)
            if istream < 0:
                break
            istart = pdf.find(startmark, istream, istream + 20)
            if istart < 0:
                i = istream + 20
                continue
            iend = pdf.find(b"endstream", istart)
            if iend < 0:
                raise Exception("Didn't find end of stream!")
            iend = pdf.find(endmark, iend - 20)
            if iend < 0:
                raise Exception("Didn't find end of JPG!")

            istart += startfix
            iend += endfix
            jpg = pdf[istart:iend]

            fp = op.join(dn, "jpg%d.jpg" % njpg)
            jpeg.append(fp)
            with open(fp, "wb") as jpgfile:
                jpgfile.write(jpg)

            njpg += 1
            i = iend
        return jpeg

    pdf = self.pdf(name)

    f, fp1 = tempfile.mkstemp(suffix='.pdf')
    os.close(f)
    log.debug('Saving it in %s.'%fp1)
    pdf.get(dest=fp1)

    jpeg = extract_file(fp1)

    if len(glob(op.join(op.dirname(fp1), '*.jpg'))) == 0:
        raise Exception('No snapshot found in report. %s %s'%(name, fp1))

    f, fp2 = tempfile.mkstemp(suffix='.jpg')
    os.close(f) # we dont need the handle
    cmd = 'montage %s/jpg*jpg -background black -geometry 2600x+0+0 '\
        '-tile 1x %s'%(op.dirname(fp1), fp2)
    os.system(cmd)

    cmd = 'mv %s %s'%(fp2, fp)
    os.system(cmd)
    for each in jpeg:
        cmd = 'rm %s'%each
        os.system(cmd)
    cmd = 'rm %s'%fp1
    os.system(cmd)



import os
from pyxnat import Interface
from . import skip_if_no_network
central = Interface(config=os.path.join(os.path.dirname(os.path.abspath(__file__)), 'central.cfg'))

notified = []

def notify(cobj, eobj):
    notified.append(eobj)

def test_register_callback():
    local_notified = []
    central.manage.register_callback(notify)
    for element in central.select('/projects'):
        local_notified.append(element)

    assert notified == local_notified

def test_unregister_callback():
    central.manage.unregister_callback()
    assert central._callback is None

@skip_if_no_network
def test_add_schema():
    assert(len(central.manage.schemas()) == 0)
    central.manage.schemas.add(url='/xapi/schemas/xnat')
    assert(list(central.manage.schemas()) == ['xnat'])



XNAT_RESOURCE_NAME = 'ASHS'

def volumes(self, mode='corr_nogray'):
    import pandas as pd

    f = list(self.files('*icv.txt'))[0]
    uri = f._uri
    res = self._intf.get(uri).text.split(' ')

    f = list(self.files('*_left_%s_volumes.txt'%mode))[0]
    uri = f._uri
    resl = self._intf.get(uri).text.split('\n')

    f = list(self.files('*_right_%s_volumes.txt'%mode))[0]
    uri = f._uri
    resr = self._intf.get(uri).text.split('\n')

    table = []
    for resx in [resl, resr]:
        for line in resx:
                if line == '':
                    continue
                line = line.split(' ')
                s = line[0]
                side = line[1]
                region = line[2]
                i = int(line[-2])
                m = float(line[-1])
                table.append([s, side, region, i, m])

    table.append([res[0], None, 'tiv', None, float(res[1].rstrip('\n'))])

    columns = ['subject', 'side', 'region', 'n_slices', 'volume']
    return pd.DataFrame(table, columns=columns)



import os.path as op
from pyxnat import Interface
from . import skip_if_no_network

fp = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
central = Interface(config=fp)
central_anon = Interface('https://central.xnat.org', anonymous=True)

def test_simple_object_listing():
    assert isinstance(central.select.projects().get(), list)

def test_simple_path_listing():
    assert isinstance(central.select('/projects').get(), list)

def test_nested_object_listing():
    assert isinstance(central.select.projects('*OASIS*').subjects().get(), list)

def test_nested_path_listing():
    assert isinstance(central.select('/projects/*OASIS*/subjects').get(), list)

@skip_if_no_network
def test_search_access():
    constraints = [('xnat:subjectData/PROJECT', '=', 'CENTRAL_OASIS_CS'), 'AND']

    for subject in central.select('//subjects').where(constraints):
        assert '/projects/CENTRAL_OASIS_CS' in subject._uri

def test_connection_with_explicit_parameters():
    import json
    cfg = json.load(open(fp))
    x = Interface(server=cfg['server'], user=cfg['user'],
        password=cfg['password'])

def test_anonymous_access():
    projects = central_anon.select.projects().get()
    assert isinstance(projects, list)
    assert list

@skip_if_no_network
def test_close_jsession():
    config_file = op.join(op.dirname(op.abspath(__file__)), 'central.cfg')
    with Interface(config=config_file) as central:
        assert central.select.project('nosetests').exists()

def test_save_config():
    central.save_config('/tmp/.xnat.cfg')

@skip_if_no_network
def test_version():
    v = central.version()
    assert(v['version'] == '1.7.5.2-SNAPSHOT')

def test_login_using_explicit_credentials():
    x = Interface(server='http://server/',
                user='user', password='password')



#!/usr/bin/env python

import setuptools
import os.path as op


def get_version():
    try:
        import re
        import os.path as op
        fp = op.join(op.dirname(__file__), 'pyxnat', 'version.py')
        s = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format('VERSION'),
                      open(fp).read())
        return s.group(1)
    except Exception:
        raise RuntimeError("No version found")


this_directory = op.abspath(op.dirname(__file__))
with open(op.join(this_directory, 'README.rst')) as f:
    long_description = f.read()


setuptools.setup(name='bbrc-pyxnat',
      version=get_version(),
      summary='XNAT in Python',
      author='Yannick Schwartz',
      author_email='yannick.schwartz@cea.fr',
      url='http://packages.python.org/pyxnat/',
      packages=setuptools.find_packages(exclude=('doc*', 'tests')),
      description='XNAT in Python',
      long_description=long_description,
      long_description_content_type='text/x-rst',
      license='BSD',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Environment :: Console',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Intended Audience :: Education',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Topic :: Scientific/Engineering',
          'Topic :: Utilities',
          'Topic :: Internet :: WWW/HTTP',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 2.7',
      ],

      platforms='any',
      scripts=['bin/sessionmirror.py'],
      install_requires=['lxml>=4.3',
        'requests>=2.20',
        'requests[security]'],
      package_data={'pyxnat': ['README.rst'], },

      )



import os
import tempfile

from .jsonutil import JsonTable, csv_to_json
from .resources import CObject


class Tags(object):
    """ Database tags sub-interface.
        It is meant to be an attribute of the main Interface class.
    """

    def __init__(self, interface):
        self._intf = interface
        self._meta_project = None

    def __call__(self):
        self._init()
        return self._meta_project.subject(self._intf._user
                                          ).resource('tags').files().get()

    def _init(self):
        if self._meta_project is None:
            self._meta_project = \
                self._intf.select.project('metabase_%s' % self._intf._user)
            if self._meta_project.accessibility() != 'private':
                self._meta_project.set_accessibility('private')

        if not self._meta_project.exists():
            self._meta_project.create()
            self._meta_project.set_accessibility('private')

    def new(self, name):
        return self.get(name).create()

    def delete(self, name):
        self.get(name).delete()

    def get(self, name):
        return Tag(name, self._intf)

    def share(self, other_user):
        if self._intf.select.project('metabase_%s' % other_user).exists():
            self._meta_project.subject(self._intf._user
                                       ).share('metabase_%s' % other_user)


class Tag(object):
    def __init__(self, name, interface):
        self._name = name
        self._intf = interface
        self._intf.manage.tags._init()
        self._file = self._intf.manage.tags._meta_project.subject(
            self._intf._user).resource('tags').file(name)

    def __repr__(self):
        return '<Tag> %s'%self._name

    def _read(self):
        fd = open(self._file.get(), 'rb')
        jtag = JsonTable(csv_to_json(fd.read()))
        fd.close()

        return jtag

    def create(self):
        if not self.exists():
            jtag = JsonTable([])
            tmp = tempfile.mkstemp()[1]
            jtag.dump_csv(tmp)
            self._file.put(tmp)
            os.remove(tmp)

        return self

    def delete(self):
        if self.exists():
            self._file.delete()

    def exists(self):
        return self._file.exists()

    def dereference(self, uri):
        jtag = self._read()
        tmp = tempfile.mkstemp()[1]
        jtag.where_not(URI=uri).dump_csv(tmp)
        self._file.put(tmp)
        os.remove(tmp)

    def dereference_many(self, uris=[]):
        jtag = self._read()
        for uri in uris:
            jtag = jtag.where_not(URI=uri)
        tmp = tempfile.mkstemp()[1]
        jtag.dump_csv(tmp)
        self._file.put(tmp)
        os.remove(tmp)

    def reference(self, uri):
        uri = self._intf.select(uri)._uri
        jtag = self._read()
        if uri not in jtag.get('URI', always_list=True):
            jtag.data.append({'URI':uri})
            tmp = tempfile.mkstemp()[1]
            jtag.dump_csv(tmp)
            self._file.put(tmp)
            os.remove(tmp)

    def reference_many(self, uris=[]):
        jtag = self._read()
        for uri in uris:
            jtag.data.append({'URI':uri})
        tmp = tempfile.mkstemp()[1]
        jtag.dump_csv(tmp)
        self._file.put(tmp)
        os.remove(tmp)

    def references(self, show_uris=False):
        jtag = self._read()
        uris = jtag.get('URI', always_list=True)

        if not show_uris:
            return CObject(uris, self._intf)
        return uris



import os
import json

# import six
# if six.PY2:
#     from urllib2 import urlopen
# elif six.PY3:
#     from urllib.request import urlopen

# import smtplib
# from copy import deepcopy
# import email.mime.text
# import xml.dom.minidom

# import suds.client
# import suds.xsd.doctor

from . import httputil


class PipelineNotFoundError(Exception):
    """workflow not found"""


class Pipelines(object):

    def __init__(self, project, interface):
        self._intf = interface
        self._project = project

    def get(self):
        response =  self._intf._exec('%s/projects/%s/pipelines' % (
                self._intf._get_entry_point(),
                self._project,
                ))

        return json.loads(response)['ResultSet']['Result']

    def add(self, location):
        f = open(location, 'rb')
        pip_doc = f.read()
        f.close()

        body, content_type = httputil.file_message(
            pip_doc, 'text/xml', location, os.path.split(location)[1])

        pipeline_uri = '%s/projects/%s/pipelines/%s' % (
            self._intf._get_entry_point(),
            self._project,
            os.path.split(location)[1]
            )

        self._intf._exec(pipeline_uri,
                         method='PUT',
                         body=body,
                         headers={'content-type':content_type}
                         )

    def delete(self, pipeline_id):
        pass


class Pipeline(object):

    def __init__(self, pipeline_id, interface):
        self._intf = interface
        self._id = pipeline_id

    def run(self):
        pass

    def stop(self):
        pass

    def update(self):
        pass

    def complete(self):
        pass

    def fail(self):
        pass

# class Pipeline(object):

#     """class for mirroring workflow information (XML) in XNAT"""

#     def __init__(self, base_url, username, password, workflow_id):
#         self._base_url = base_url
#         self._username = username
#         self._password = password
#         self._cookiejar = None
#         res = self._call('CreateServiceSession.jws',
#                          'execute',
#                          (),
#                          authenticated=True)
#         self._session = str(res)
#         args = (('ns1:string', self._session),
#                 ('ns1:string', 'wrk:workflowData.ID'),
#                 ('ns1:string', '='),
#                 ('ns1:string', workflow_id),
#                 ('ns1:string', 'wrk:workflowData'))
#         workflow_ids = self._call('GetIdentifiers.jws', 'search', args)
#         self._doc = None
#         for w_id in workflow_ids:
#             url = '%s/app/template/XMLSearch.vm/id/%s/data_type/wrk:workflowData' % (self._base_url, str(w_id))
#             r = urllib2.Request(url)
#             self._cookiejar.add_cookie_header(r)
#             data = urllib2.urlopen(r).read()
#             doc = xml.dom.minidom.parseString(data)
#             workflow_node = doc.getElementsByTagName('wrk:Workflow')[0]
#             status = workflow_node.getAttribute('status').lower()
#             if status in ('queued', 'awaiting action', 'hold'):
#                 self._doc = doc
#                 break
#         if self._doc is None:
#             raise PipelineNotFoundError
#         return

#     def _call(self,
#               jws,
#               operation,
#               inputs,
#               authenticated=False,
#               fix_import=False):
#         """perform a SOAP call"""
#         url = '%s/axis/%s' % (self._base_url, jws)
#         if authenticated:
#             t = suds.transport.http.HttpAuthenticated(username=self._username,
#                                                       password=self._password)
#         else:
#             t = suds.transport.http.HttpTransport()
#         if self._cookiejar is not None:
#             t.cookiejar = self._cookiejar
#         if fix_import:
#             xsd_url = 'http://schemas.xmlsoap.org/soap/encoding/'
#             imp = suds.xsd.doctor.Import(xsd_url)
#             doctor = suds.xsd.doctor.ImportDoctor(imp)
#             client = suds.client.Client('%s?wsdl' % url,
#                                         transport=t,
#                                         doctor=doctor)
#         else:
#             client = suds.client.Client('%s?wsdl' % url, transport=t)
#         typed_inputs = []
#         for (dtype, val) in inputs:
#             ti = client.factory.create(dtype)
#             ti.value = val
#             typed_inputs.append(ti)
#         # the WSDL returns the local IP address in the URLs; these need
#         # to be corrected if XNAT is behind a proxy
#         client.set_options(location=url)
#         f = getattr(client.service, operation)
#         result = f(*typed_inputs)
#         if self._cookiejar is None:
#             self._cookiejar = t.cookiejar
#         return result


#     def _close(self):
#         """close the XNAT session (log out)"""
#         self._call('CloseServiceSession.jws', 'execute', ())
#         return

#     def _update_xnat(self):
#         """update XNAT with the current state of this (WorkflowInfo) object"""
#         inputs = (('ns0:string', self._session),
#                   ('ns0:string', self._doc.toxml()),
#                   ('ns0:boolean', False),
#                   ('ns0:boolean', True))
#         self._call('StoreXML.jws',
#                    'store',
#                    inputs,
#                    authenticated=True,
#                    fix_import=True)
#         return

#     def _append_node(self, root, name, value):
#         """add a simple text node with tag "name" and data "value" under
#         the node "root"
#         """
#         node = self._doc.createElement(name)
#         node.appendChild(self._doc.createTextNode(value))
#         root.appendChild(node)
#         return

#     def set_environment(self, arguments, parameters):
#         """set the execution environment

#         should be run only once before update() is called
#         """
#         # order is important
#         workflow_node = self._doc.getElementsByTagName('wrk:Workflow')[0]
#         ee_node = self._doc.createElement('wrk:executionEnvironment')
#         ee_node.setAttribute('xsi:type', 'wrk:xnatExecutionEnvironment')
#         workflow_node.appendChild(ee_node)
#         self._append_node(ee_node, 'wrk:pipeline', arguments['pipeline'])
#         self._append_node(ee_node, 'wrk:xnatuser', arguments['u'])
#         self._append_node(ee_node, 'wrk:host', arguments['host'])
#         params_node = self._doc.createElement('wrk:parameters')
#         ee_node.appendChild(params_node)
#         for key in parameters:
#             param_node = self._doc.createElement('wrk:parameter')
#             param_node.setAttribute('name', key)
#             for val in parameters[key]:
#                 param_node.appendChild(self._doc.createTextNode(val))
#             params_node.appendChild(param_node)
#         for email in arguments['notify_emails']:
#             self._append_node(ee_node, 'wrk:notify', email)
#         self._append_node(ee_node, 'wrk:dataType', arguments['dataType'])
#         self._append_node(ee_node, 'wrk:id', arguments['id'])
#         if arguments['notify_flag']:
#             self._append_node(ee_node, 'wrk:supressNotification', '0')
#         else:
#             self._append_node(ee_node, 'wrk:supressNotification', '1')
#         return

#     def update(self, step_id, step_description, percent_complete):
#         """update the workflow in XNAT"""
#         workflow_node = self._doc.getElementsByTagName('wrk:Workflow')[0]
#         workflow_node.setAttribute('status', 'Running')
#         t = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
#         workflow_node.setAttribute('current_step_launch_time', t)
#         workflow_node.setAttribute('current_step_id', step_id)
#         workflow_node.setAttribute('step_description', step_description)
#         workflow_node.setAttribute('percentageComplete', percent_complete)
#         self._update_xnat()
#         return

#     def complete(self):
#         """mark the workflow comleted in XNAT and close the session"""
#         workflow_node = self._doc.getElementsByTagName('wrk:Workflow')[0]
#         workflow_node.setAttribute('status', 'Complete')
#         t = datetime.datetime.now().strftime('%Y-%m-%dT%H:%M:%S')
#         workflow_node.setAttribute('current_step_launch_time', t)
#         workflow_node.setAttribute('percentageComplete', '100.0')
#         try:
#             workflow_node.removeAttribute('current_step_id')
#         except xml.dom.NotFoundErr:
#             pass
#         try:
#             workflow_node.removeAttribute('step_description')
#         except xml.dom.NotFoundErr:
#             pass
#         self._update_xnat()
#         self._close()
#         return

#     def fail(self, description=None):
#         """mark the workflow failed in XNAT and close the session"""
#         workflow_node = self._doc.getElementsByTagName('wrk:Workflow')[0]
#         workflow_node.setAttribute('status', 'Failed')
#         if description is not None:
#             workflow_node.setAttribute('step_description', description)
#         self._update_xnat()
#         self._close()
#         return



import tempfile
from uuid import uuid1
from six import string_types
import os.path as op
import os
from pyxnat import Interface
from . import skip_if_no_network
from nose import SkipTest
from pyxnat.core import interfaces

_modulepath = op.dirname(op.abspath(__file__))

central = Interface(config=op.join(op.dirname(op.abspath(__file__)),
                                   'central.cfg'))
interfaces.STUBBORN = True

_id_set1 = {
    'sid': uuid1().hex,
    'eid': uuid1().hex,
    'aid': uuid1().hex,
    'cid': uuid1().hex,
    'rid': uuid1().hex,
    }

_id_set2 = {
    'sid': uuid1().hex,
    'eid': uuid1().hex,
    'aid': uuid1().hex,
    'cid': uuid1().hex,
    'rid': uuid1().hex,
    }
subj_1 = central.select.project('nosetests3').subject(_id_set1['sid'])
expe_1 = subj_1.experiment(_id_set1['eid'])
asse_1 = expe_1.assessor(_id_set1['aid'])
scan_1 = expe_1.scan(_id_set1['cid'])
reco_1 = expe_1.reconstruction(_id_set1['rid'])


@skip_if_no_network
def test_01_subject_create():
    assert not subj_1.exists()
    subj_1.create()
    assert subj_1.exists()


@skip_if_no_network
def test_02_experiment_create():
    assert not expe_1.exists()
    expe_1.create()
    assert expe_1.exists()
    expe_1.trigger(fix_types=True, pipelines=True, scan_headers=False)


@skip_if_no_network
def test_03_assessor_create():
    assert not asse_1.exists()
    asse_1.create()
    assert asse_1.exists()


@skip_if_no_network
def test_04_scan_create():
    assert not scan_1.exists()
    scan_1.create()
    assert scan_1.exists()


@skip_if_no_network
def test_05_reconstruction_create():
    assert not reco_1.exists()
    reco_1.create()
    assert reco_1.exists()


# def test_provenance():
#     reco_1.provenance.set({'program':'nosetests'})
#     assert reco_1.provenance.get()[0]['program'] == 'nosetests'

@skip_if_no_network
def test_06_multi_create():
    asse_2 = central.select('/projects/nosetests3/subjects/%(sid)s'
                            '/experiments/%(eid)s'
                            '/assessors/%(aid)s' % _id_set2
                            )

    expe_2 = central.select('/projects/nosetests3/subjects/%(sid)s'
                            '/experiments/%(eid)s' % _id_set2
                            )

    assert not asse_2.exists()
    asse_2.create(experiments='xnat:petSessionData',
                  assessors='xnat:petAssessorData')
    assert asse_2.exists()

    assert asse_2.datatype() == 'xnat:petAssessorData'
    assert expe_2.datatype() == 'xnat:petSessionData'

    scan_2 = central.select('/projects/nosetests3/subjects/%(sid)s'
                            '/experiments/%(eid)s/scans/%(cid)s' % _id_set2
                            ).create()

    assert scan_2.datatype() == 'xnat:mrScanData'


# def test_share_subject():
#    assert not central.select('/projects/nosetests2'
#                              '/subjects/%(sid)s' % _id_set1
#                              ).exists()
#    subj_1.share('nosetests2')
#    assert central.select('/projects/nosetests2/subjects/%(sid)s'%_id_set1
#                              ).exists()
#    assert set(subj_1.shares().get()) == set(['nosetests', 'nosetests2'])

# def test_share_experiment():
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s'%_id_set1
#                          ).exists()
#    expe_1.share('nosetests2')
#    assert central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s'%_id_set1
#                          ).exists()
#    assert set(expe_1.shares().get()) == set(['nosetests', 'nosetests2'])

# def test_share_assessor():
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                              '/experiments/%(eid)s/assessors/%(aid)s'%_id_set1
#                              ).exists()
#    asse_1.share('nosetests2')
#    assert central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s/assessors/%(aid)s'%_id_set1
#                          ).exists()
#    assert set(asse_1.shares().get()) == set(['nosetests', 'nosetests2'])

# def test_unshare_assessor():
#    asse_1.unshare('nosetests2')
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                              '/experiments/%(eid)s/assessors/%(aid)s'%_id_set1
#                              ).exists()
#    assert asse_1.shares().get() == ['nosetests']

# def test_unshare_experiment():
#    expe_1.unshare('nosetests2')
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'
#                          '/experiments/%(eid)s'%_id_set1
#                          ).exists()
#    assert expe_1.shares().get() == ['nosetests']

# def test_unshare_subject():
#    subj_1.unshare('nosetests2')
#    assert not central.select('/projects/nosetests2/subjects/%(sid)s'%_id_set1
#                              ).exists()
#    assert subj_1.shares().get() == ['nosetests']

@skip_if_no_network
def test_07_put_file():
    local_path = op.join(_modulepath, 'hello_xnat.txt')
    f = subj_1.resource('test').file('hello.txt')
    subj_1.resource('test').file('hello.txt').put(local_path)
    subj_1.resource('test').put([local_path])
    assert f.exists()
    assert int(f.size()) == os.stat(local_path).st_size


@skip_if_no_network
def test_08_get_file():
    fh = subj_1.resource('test').file('hello.txt')

    fpath = fh.get()
    assert op.exists(fpath)
    print(['toto3', open(fpath, 'rb').read()])
    try:
        assert open(fpath, 'rb').read() == bytes('Hello XNAT!%s' % os.linesep,
                                                 encoding='utf8')
    except TypeError:
        pass

    custom = op.join(tempfile.gettempdir(), uuid1().hex)

    fh.get(custom)
    assert op.exists(custom), "fpath: %s custom: %s" % (fpath, custom)
    os.remove(custom)
    os.remove(fpath)


@skip_if_no_network
def test_09_put_dir_file():
    local_path = op.join(_modulepath, 'hello_again.txt')
    subj_1.resource('test').file('dir/hello.txt').put(local_path)
    assert subj_1.resource('test').file('dir/hello.txt').exists()
    assert int(subj_1.resource('test').file('dir/hello.txt').size()) == \
                                                os.stat(local_path).st_size


@skip_if_no_network
def test_10_get_dir_file():
    fh = subj_1.resource('test').file('dir/hello.txt')

    fpath = fh.get()
    assert op.exists(fpath)
    try:
        assert open(fpath, 'rb').read() == bytes('Hello again!%s' % os.linesep,
                                                 encoding='utf8')
    except TypeError:
        pass

    custom = op.join(tempfile.gettempdir(), uuid1().hex)

    fh.get(custom)
    assert op.exists(custom), "fpath: %s custom: %s" % (fpath, custom)
    os.remove(custom)
    os.remove(fpath)


@skip_if_no_network
def test_11_get_copy_file():
    fpath = op.join(tempfile.gettempdir(), uuid1().hex)
    fpath = subj_1.resource('test').file('hello.txt').get_copy(fpath)
    assert op.exists(fpath)
    fd = open(fpath, 'rb')
    try:
        assert fd.read() == bytes('Hello XNAT!%s' % os.linesep,
                                                    encoding='utf8')
    except TypeError:
        pass
    fd.close()
    os.remove(fpath)


@skip_if_no_network
def test_12_file_last_modified():
    f = subj_1.resource('test').file('hello.txt')
    assert isinstance(f.last_modified(), string_types)
    assert len(f.last_modified()) > 0
    f.delete()
    assert(not f.exists())
    r = subj_1.resource('test')
    r.delete()
    assert(not r.exists())


@skip_if_no_network
def test_13_last_modified():
    sid = subj_1.id()

    t1 = central.select('/project/nosetests3').last_modified()[sid]
    subj_1.attrs.set('age', '26')
    assert subj_1.attrs.get('age') == '26'
    t2 = central.select('/project/nosetests3').last_modified()[sid]

    assert t1 != t2


@skip_if_no_network
def test_14_getitem_key():
    projects = central.select.projects()
    assert projects.first().id() == projects[0].id()
    piter = projects.__iter__()
    next(piter)
    assert next(piter).id() == projects[1].id()


@skip_if_no_network
def test_15_getitem_slice():
    projects = central.select.projects()
    assert projects.first().id() == next(projects[:1]).id()
    piter = projects.__iter__()
    next(piter)
    next(piter)
    next(piter)
    for pobj in projects[3:6]:
        assert next(piter).id() == pobj.id()


def test_subject1_parent():
    project = central.select.project('nosetests3')
    assert subj_1.parent()._uri == project._uri


def test_project_parent():
    project = central.select.project('nosetests3')
    assert not project.parent()


@skip_if_no_network
def test_16_subject1_delete():
    assert subj_1.exists()
    subj_1.delete()
    assert not subj_1.exists()


@skip_if_no_network
def test_17_subject2_delete():
    subj_2 = central.select('/projects/nosetests3/subjects/%(sid)s'%_id_set2)
    assert subj_2.exists()
    subj_2.delete()
    assert not subj_2.exists()


@skip_if_no_network
def test_18_project_configuration():
    project = central.select('/project/nosetests3')
    version = central.version()
    from pyxnat.core.errors import DatabaseError

    try:
        assert project.quarantine_code() == 0
        assert project.prearchive_code() == 4, project.prearchive_code()
    except DatabaseError:
        if version['version'] == '1.7.5.2-SNAPSHOT':
            msg = 'Version 1.7.5.2-SNAPSHOT gives trouble on some machines. \
                   Skipping it'
            raise SkipTest(msg)

    if version['version'] != '1.7.5.2-SNAPSHOT':
        try:
            assert project.current_arc() == b'arc001'
        except DatabaseError:
            msg = 'Check if current_arc is supported in XNAT version %s.'\
                %version['version']
            print(msg)


    assert 'nosetests' in project.users()
    assert 'nosetests' in project.owners()
    assert project.user_role('nosetests') == 'owner'


@skip_if_no_network
def test_19_put_zip():
    local_path = op.join(_modulepath, 'hello_dir.zip')
    assert op.exists(local_path)

    # Upload and confirm proper extraction
    r1 = subj_1.resource('test_zip_extract')
    r1.put_zip(local_path, extract=True)
    assert r1.exists()
    assert r1.file('hello_dir/hello_xnat_dir.txt').exists()
    assert r1.file('hello_dir/hello_dir2/hello_xnat_dir2.txt').exists()

    # Upload and confirm not extracted
    r2 = subj_1.resource('test_zip_no_extract')
    r2.put_zip(local_path, extract=False)
    assert r2.exists()
    assert r2.file('hello_dir.zip').exists()


@skip_if_no_network
def test_20_get_zip():
    r = subj_1.resource('test_zip_extract')
    local_dir = op.join(_modulepath, 'test_zip_download'+r.id())
    file_list = ['test_zip_extract/hello_dir',
                 'test_zip_extract/hello_dir/hello_xnat_dir.txt',
                 'test_zip_extract/hello_dir/hello_dir2',
                 'test_zip_extract/hello_dir/hello_dir2/hello_xnat_dir2.txt']
    file_list = [op.join(local_dir, e) for e in file_list]

    if not op.exists(local_dir):
        os.mkdir(local_dir)

    r.get(local_dir, extract=True)
    for f in file_list:
        assert op.exists(f)
    r.get(local_dir, extract=False)
    assert op.isfile(op.join(local_dir, 'test_zip_extract.zip'))


@skip_if_no_network
def test_21_project_aliases():
    project = central.select('/project/nosetests3')
    assert project.aliases() ==  ['nosetests32']


@skip_if_no_network
def test_22_project():
    project = central.select.project('nosetests3')
    project.datatype()
    project.experiments()
    project.experiment('nose')
